
#include "malloc.h"
#include "as31glue.h"
#include "printf.h"


/* this "simm_id" type is what will normally be passed around, */
/* and declared inside of other structs (like playlists, file */
/* lists, etc.  You can't actually access memory with it until */
/* you pass it through a function to turn it into a proper C */
/* xdata pointer (see below) */

//  defined in malloc.h
// typedef unsigned long  simm_id;


/* Normally this complex union won't be used */
/* This funny union represents a block/offset memory address */
/* and allows it to be passed between functions as a 32 bit */
/* long, because SDCC can't pass structs or unions. */

// typedef union dram_memory_pointer_struct {
typedef union {
	struct {
		unsigned int offset;
		unsigned int block;
	} part;
	simm_id addr32;
} simm_id_union;


#define DATA_PER_BLOCK 4090


struct simm_block_struct {
	unsigned int next;
	unsigned int free_bytes;
	unsigned int ref_count;
	unsigned char bytes[DATA_PER_BLOCK];
};



static data unsigned int first_malloced_block=0;
static data unsigned int last_malloced_block;


/* don't try this sort of direct hardware access at home! */
volatile static xdata at 0xFF0E unsigned int block7000;
static xdata at 0x7000 struct simm_block_struct p7;


simm_id simm_malloc(unsigned int num_bytes)
{
	unsigned int block;
	simm_id_union mem;

	printf("simm_malloc begin, f=%d\r\n", first_malloced_block);

	if (num_bytes > DATA_PER_BLOCK) return 0;

	if (first_malloced_block == 0) {
		block7000 = first_malloced_block = \
		   last_malloced_block = malloc_blocks(1);
		printf("malloc 1st: %d\r\n", block7000);
		p7.next = 0;
		p7.free_bytes = DATA_PER_BLOCK;
		p7.ref_count = 0;
	} else {
		block7000 = last_malloced_block;
		if (p7.free_bytes < num_bytes) {
			block = malloc_blocks(1);
			printf("malloc Nth: %d\r\n", block7000);
			p7.next = last_malloced_block = block;
			block7000 = block;
			p7.next = 0;
			p7.free_bytes = DATA_PER_BLOCK;
			p7.ref_count = 0;
		}
	}
	mem.part.block = block7000;
	mem.part.offset = 4096 - p7.free_bytes;
	p7.free_bytes -= num_bytes;
	p7.ref_count++;
	return mem.addr32;
}


void simm_free(simm_id addr32)
{
	unsigned int block, next_block;
	simm_id_union mem;

	mem.addr32 = addr32;
	block7000 = mem.part.block;

	if (--(p7.ref_count) != 0) return;

	if (mem.part.block == first_malloced_block) {
		first_malloced_block = p7.next;
		free_blocks(mem.part.block);
		return;
	}

	next_block = p7.next;
	free_blocks(mem.part.block);

	for (block = first_malloced_block; block != 0; block = p7.next) {
		block7000 = block;
		if (p7.next == mem.part.block) {
			p7.next = next_block;
			return;
		}
	}

	printf("Error: could find freed block\r\n");
}






/* This is the magic function that takes the "pointer" type */
/* from malloc and maps the referenced block into the 8051's */
/* address space and returns a SDCC xdata * to the memory */
/* See http://www.pjrc.com/tech/mp3/mem_map.html for more */
/* details about how the DRAM controller page allocation works */

#define DRAM_PAGE_CFG 0xFF00

xdata void * addr6(simm_id addr32)
{
	addr32;	 /* suppress unused variable warning */

	_asm
	mov	r2, dpl		;put offset into r2/r3
	mov	r3, dph
	mov	r4, a		;keep msb of block in r4
	mov	dptr, #DRAM_PAGE_CFG + (6 * 2)	;always map in 7000-7FFF
	mov	a, b
	movx	@dptr, a
	inc	dptr
	mov	a, r4
	movx	@dptr, a	;map the block to 0x7000
	mov	dpl, r2
	mov	a, r3
	anl	a, #15
	add	a, #0x60
	mov	dph, a		;now dptr points to the users data
	_endasm;
}


xdata void * addr7(simm_id addr32)
{
	addr32;	 /* suppress unused variable warning */

	_asm
	mov	r2, dpl		;put offset into r2/r3
	mov	r3, dph
	mov	r4, a		;keep msb of block in r4
	mov	dptr, #DRAM_PAGE_CFG + (7 * 2)	;always map in 7000-7FFF
	mov	a, b
	movx	@dptr, a
	inc	dptr
	mov	a, r4
	movx	@dptr, a	;map the block to 0x7000
	mov	dpl, r2
	mov	a, r3
	anl	a, #15
	add	a, #0x70
	mov	dph, a		;now dptr points to the users data
	_endasm;
}

